"""
If "Draw Arrows" is selected, arrows will be drawn at the end of each line segment, the default is on.

"""

from __future__ import absolute_import
#Init has to be imported first because it has code to workaround the python bug where relative imports don't work if the module is imported as a main module.
import __init__

from skeinforge_tools.analyze_plugins.analyze_utilities import tableau
from skeinforge_tools.skeinforge_utilities import preferences

__author__ = "Enrique Perez (perez_enrique@yahoo.com)"
__date__ = "$Date: 2008/21/04 $"
__license__ = "GPL 3.0"


def getConstructor( window ):
	"Get the repository constructor."
	return ViewpointMove( window )


class ViewpointMove( tableau.MouseToolBase ):
	"Display the line when it is clicked."
	def __init__( self, window ):
		"Set the canvas and window."
		self.setCanvasItems( window.canvas )
		self.buttonOnePressedScreenCoordinate = None
		self.window = window

	def button1( self, event, shift = False ):
		"Print line text and connection line."
		self.destroyEverything()
		self.buttonOnePressedScreenCoordinate = complex( event.x, event.y )
		self.scrollPaneCenter = self.window.getScrollPaneCenter()

	def buttonRelease1( self, event, shift = False ):
		"The left button was released, <ButtonRelease-1> function."
		self.destroyEverything()

	def destroyEverything( self ):
		"Destroy items."
		self.buttonOnePressedScreenCoordinate = None
		self.destroyItems()

	def motion( self, event, shift = False ):
		"The mouse moved, <Motion> function."
		if self.buttonOnePressedScreenCoordinate == None:
			return
		motionCoordinate = complex( event.x, event.y )
		relativeMotionCoordinate = motionCoordinate - self.buttonOnePressedScreenCoordinate
		relativeScrollPaneMotionX = - relativeMotionCoordinate.real / float( self.window.screenSize.real )
		relativeScrollPaneMotionY = - relativeMotionCoordinate.imag / float( self.window.screenSize.imag )
		self.window.relayXview( preferences.Tkinter.MOVETO, self.scrollPaneCenter.real * self.window.oneMinusCanvasWidthOverScreenWidth + relativeScrollPaneMotionX )
		self.window.relayYview( preferences.Tkinter.MOVETO, self.scrollPaneCenter.imag * self.window.oneMinusCanvasHeightOverScreenHeight + relativeScrollPaneMotionY )
