"""
Help is a script to access the plugins which help a gcode file.

The plugin buttons which are commonly used are bolded and the ones which are rarely used have normal font weight.

		self.helpMenu.add_cascade( label = "Forums", menu = self.forumsMenu )
		self.forumsMenu.add_command( label = 'Bits from Bytes Printing', command = HelpPage().getOpenFromAfterWWW( 'bitsfrombytes.com/fora/user/index.php?board=5.0' ) )
		self.forumsMenu.add_command( label = 'Bits from Bytes Software', command = HelpPage().getOpenFromAfterWWW( 'bitsfrombytes.com/fora/user/index.php?board=4.0' ) )
		self.forumsMenu.add_separator()
		self.forumsMenu.add_command( label = 'Skeinforge Contributions', command = HelpPage().getOpenFromAfterHTTP( 'dev.forums.reprap.org/read.php?12,27562' ) )
		self.forumsMenu.add_command( label = 'Skeinforge Powwow', command = HelpPage().getOpenFromAfterHTTP( 'dev.forums.reprap.org/read.php?12,20013' ) )
		self.forumsMenu.add_command( label = 'Skeinforge Settings', command = HelpPage().getOpenFromAfterHTTP( 'dev.forums.reprap.org/read.php?12,27434' ) )
		self.helpMenu.add_command( label = 'Index', command = HelpPage().getOpenFromDocumentationSubName( '' ) )
		self.helpMenu.add_command( label = 'Manual', command = HelpPage().getOpenFromAfterWWW( 'bitsfrombytes.com/wiki/index.php?title=Skeinforge' ) )
		self.helpMenu.add_command( label = 'Overview', command = HelpPage().getOpenFromDocumentationSubName( 'skeinforge.html' ) )
"""

from __future__ import absolute_import
#Init has to be imported first because it has code to workaround the python bug where relative imports don't work if the module is imported as a main module.
import __init__

from skeinforge_tools.skeinforge_utilities import gcodec
from skeinforge_tools.skeinforge_utilities import preferences
import os


__author__ = "Enrique Perez (perez_enrique@yahoo.com)"
__date__ = "$Date: 2008/21/04 $"
__license__ = "GPL 3.0"


def addToMenu( menu, repository ):
	"Add a tool plugin menu."
	path = preferences.getPathFromFileNameHelp( repository.fileNameHelp )
	menu.add_command( label = os.path.basename( path ).capitalize(), command = preferences.HelpPage().getOpenFromDocumentationSubName( repository.fileNameHelp ) )
	menu.add_separator()
	helpRepository = HelpRepository()
	preferences.addMenuEntitiesToMenu( menu, helpRepository.menuEntities )

def getRepositoryConstructor():
	"Get the repository constructor."
	return HelpRepository()


class HelpRepository:
	"A class to handle the help preferences."
	def __init__( self ):
		"Set the default preferences, execute title & preferences fileName."
		#Set the default preferences.
		preferences.addListsToRepository( self )
		preferences.LabelDisplay().getFromName( 'Index of Local Documentation: ', self )
		preferences.HelpPage().getFromNameSubName( 'Index', self )
		preferences.LabelDisplay().getFromName( 'Manual, Wiki with Pictures & Charts: ', self )
		preferences.HelpPage().getFromNameAfterWWW( 'bitsfrombytes.com/wiki/index.php?title=Skeinforge', 'Manual', self )
		preferences.LabelDisplay().getFromName( 'Overview of Skeinforge: ', self )
		preferences.HelpPage().getFromNameSubName( 'Overview', self, 'skeinforge.html' )
		preferences.LabelSeparator().getFromRepository( self )
		preferences.LabelDisplay().getFromName( 'Forums:', self )
		preferences.LabelDisplay().getFromName( 'Bits from Bytes Printing Board:', self )
		preferences.HelpPage().getFromNameAfterWWW( 'bitsfrombytes.com/fora/user/index.php?board=5.0', 'Bits from Bytes Printing Board', self )
		preferences.LabelDisplay().getFromName( 'Bits from Bytes Software Board:', self )
		preferences.HelpPage().getFromNameAfterWWW( 'bitsfrombytes.com/fora/user/index.php?board=4.0', 'Bits from Bytes Software Board', self )
		preferences.LabelDisplay().getFromName( 'Skeinforge Contributions Thread:', self )
		preferences.HelpPage().getFromNameAfterHTTP( 'dev.forums.reprap.org/read.php?12,27562', 'Skeinforge Contributions Thread', self )
		preferences.LabelDisplay().getFromName( 'Skeinforge Powwow Thread:', self )
		preferences.HelpPage().getFromNameAfterHTTP( 'dev.forums.reprap.org/read.php?12,20013', 'Skeinforge Powwow Thread', self )
		preferences.LabelDisplay().getFromName( 'Skeinforge Settings Thread:', self )
		preferences.HelpPage().getFromNameAfterHTTP( 'dev.forums.reprap.org/read.php?12,27434', 'Skeinforge Settings Thread', self )
		#Create the archive, title of the execute button, title of the dialog & preferences fileName.
		self.executeTitle = None
		preferences.setHelpPreferencesFileNameTitleWindowPosition( self, 'skeinforge_tools.help.html' )


def main():
	"Display the help dialog."
	preferences.startMainLoopFromConstructor( getRepositoryConstructor() )

if __name__ == "__main__":
	main()
